/**
 * File for load simulation logic.
 */


// User Import Headers.
#ifndef user_headers
    #define user_headers
    #include "simulate_loads.h"
    #include "structs.h"
    #include "terminal_commands.h"
#endif


void display_status(thread_struct* thread_args_ptr, int process_num, int process_load_status, int init_run) {
    int index;

    // printf("init_run: %i\n", init_run);

    // Travel to desired terminal row. Skip if first loop through thread displaying.
    if (init_run == 0) {
        for (index = 0; index < (thread_args_ptr->total_processors - process_num); index ++) {
            terminal_line_up();
        }
    }
    // Clear row.
    terminal_line_erase();
    terminal_line_start();

    // Display message.
    printf("Thread %5i: %5i / %5i Loads Remaining\n", process_num, process_load_status, thread_args_ptr->total_loads);

    // Travel back to original terminal location. Skip if first loop through thread displaying.
    if (init_run == 0) {
        for (index = 0; index < (thread_args_ptr->total_processors - process_num); index ++) {
            terminal_line_down();
        }
        terminal_line_start();
    }
}


/**
 * Logic to run "Asynchronous Round Robin" load scheme.
 */
void run_arr(thread_struct* thread_args_ptr) {
    int process_rank = thread_args_ptr->thread_num;

    if (process_rank == 0) {
        printf("Starting ARR load scheme.\n");
        printf("\n");
    }

    // Wait for all processes to synchronize.
    // Also sleep for a second to allow time for console output to process.
    MPI_Barrier(MPI_COMM_WORLD);
    sleep(1);

    // Handle based on process number.
    if (process_rank > 0) {
        // Child process. Handles all the work.
        worker_arr(thread_args_ptr);
    } else {
        // Main process. Handles communication to terminal.
        main_arr(thread_args_ptr);
    }

    // Wait for all processes to synchronize.
    // Also sleep for a second to allow time for console output to process.
    MPI_Barrier(MPI_COMM_WORLD);
    sleep(1);

    if (process_rank == 0) {
        printf("\n");
        printf("ARR load scheme. Finished.\n");
    }
}


/**
 * Main processor logic for "Asynchronous Round Robin" load scheme.
 */
void main_arr(thread_struct* thread_args_ptr) {
    int index = 1;
    int main_loop_bool = thread_args_ptr->total_loads;
    int init_run = 1;

    // Get messages from all worker processors.
    while (main_loop_bool >= 0) {

        int* recv_array = calloc(3, sizeof(int));
        MPI_Recv(recv_array, 2, MPI_INT, index, MPI_ANY_TAG, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
        display_status(thread_args_ptr, recv_array[0], recv_array[1], init_run);
        free(recv_array);

        // Increment for next thread.
        index = (index + 1) % thread_args_ptr->total_processors;

        // Check if past last processor.
        if (index == 0) {
            index += 1;

            init_run = 0;
            main_loop_bool -= 1;
            sleep(1);
        }

    }
}


/**
 * Worker processor logic for "Asynchronous Round Robin" load scheme.
 */
void worker_arr(thread_struct* thread_args_ptr) {
    int process_rank = thread_args_ptr->thread_num;
    int load_num = thread_args_ptr->total_loads;

    // Loop all worker processes until load count message of "0" has been sent.
    while (load_num >= 0) {

        // Sent message for current load.
        int* send_array = calloc(2, sizeof(int));
        send_array[0] = process_rank;
        send_array[1] = load_num;
        MPI_Send(send_array, 2, MPI_INT, 0, 0, MPI_COMM_WORLD);
        free(send_array);

        // Decriment load number.
        load_num -= 1;
    }
}


/**
 * Logic to run "Global Round Robin" load scheme.
 */
void run_grr(int total_processors) {
    printf("Starting GRR load scheme.");
    printf("\n");

    // pthread_t* thread_pool = NULL;

    printf("\n");
    printf("GRR load scheme. Finished.");
}


/**
 * Logic to run "Random Poling" load scheme.
 */
void run_rp(int total_processors) {
    printf("Starting RP load scheme.");
    printf("\n");

    // pthread_t* thread_pool = NULL;

    printf("\n");
    printf("RP load scheme. Finished.");
}



/**
 * Logic to run "Nearest Neighbor" load scheme.
 */
void run_nn(int total_processors) {
    printf("Starting NN load scheme.");
    printf("\n");

    // pthread_t* thread_pool = NULL;

    printf("\n");
    printf("NN load scheme. Finished.");
}


// /**
//  * Function to simulate working on a load.
//  */
// void simulate_load(char* thread_name) {
//     printf("Simulating load processing on thread \"%s\".\n", thread_name);

//     int load_counter = 0;
//     int index = 0;

//     // Iterate through loads. Each load has indexes to simulate "work".
//     while (load_counter < total_loads) {

//         // Increment load num.
//         load_counter += 1;
//         printf("Simulating load %i on thread \"%s\".\n", load_counter, thread_name);

//         // Simulate running the load.
//         index = 0;
//         while (index < indexes_per_load) {

//             // Print out index value.
//             if ((index % 10) == 0) {
//                 printf("load %i / %i    index %i / %i\n", load_counter, total_loads, index, indexes_per_load);
//             }

//             // Wait 1 second to simulate work.
//             sleep(1);

//             // Increment index.
//             index += 1;
//         }

//     }

// }


/**
 * Function to simulate working on an ARR load.
 */
// void* simulate_arr_load(thread_struct* thread_struct_ptr) {
//     printf("Simulating load processing on thread \"%i\".\n", thread_struct_ptr->thread_num);

//     // Send message back to main thread.
//     // char* message = "Thread %i checking in!\n";
//     // MPI_Send(message, sizeof(message) + 1, MPI_CHAR, 0, 0,)

//     // int load_counter = 0;
//     // int index = 0;

//     // // Iterate through loads. Each load has indexes to simulate "work".
//     // while (load_counter < total_loads) {

//     //     // Increment load num.
//     //     load_counter += 1;
//     //     printf("Simulating load %i on thread \"%s\".\n", load_counter, thread_name);

//     //     // Simulate running the load.
//     //     index = 0;
//     //     while (index < indexes_per_load) {

//     //         // Print out index value.
//     //         if ((index % 10) == 0) {
//     //             printf("load %i / %i    index %i / %i\n", load_counter, total_loads, index, indexes_per_load);
//     //         }

//     //         // Wait 1 second to simulate work.
//     //         sleep(1);

//     //         // Increment index.
//     //         index += 1;
//     //     }

//     // }

//     free(thread_struct_ptr);
// }

